  (* +--------------------------------------------------------------------+
     |                                                                    |
     |                valuation d'une formule avec priorit              |
     |                                                                    |
     |  M. Quercia, mars 1996                                             |
     +--------------------------------------------------------------------+ *)


type 'a token =
| Valeur   of 'a                      (* valeur reconnues                *)
| Prfixe  of 'a -> 'a                (* oprateur prfixe               *)
| Postfixe of 'a -> 'a                (* oprateur postfixe              *)
| Infixe   of int * ('a -> 'a -> 'a)  (* priorit, oprateur infixe      *)
| Liste    of 'a list -> 'a           (* oprateur prfixe sur une liste *)
| Lpar     			      (* parenthse gauche		 *)
| Rpar     			      (* parenthse droite		 *)
| Lsep                                (* sparateur de liste             *)
;;

exception Erreur of exn * int;;

value analyse : ('a -> 'b token) -> (char stream -> 'a stream) -> string -> 'b;;

(* [analyse f lexer s] analyse la chaine [s] et renvoie la valeur qu'elle
   reprsente. [s] est d'abord convertie en flux de caractres, ce flux est
   soumis  l'analyseur lexical [lexer], puis les lexmes obtenus sont
   convertis en lments syntaxiques par [f].
   L'analyseur proprement dit value le flux d'lments syntaxiques dans
   l'ordre suivant :

     - valuation des expressions entre parenthses ;
     - valuation des oprateurs prfixes ;
     - valuation des oprateurs postfixes ;
     - valuation des oprateurs infixes par priorit dcroissante,
       a op b op' c est valu comme (a op b) op' c lorsque op et op'
       ont mme priorit.

   Le code de l'analyseur suppose que tous les oprateurs infixe ont une
   priorit suprieure ou gale  1, la priorit 0 est rserve pour usage
   interne.

   En cas d'erreur de syntaxe ou de calcul, il est gnr une exception
   [Erreur(ex,i)] o [ex] est l'exception dclenche, et [i] est la position
   du caractre ayant dclench cette exception.

   Par exemple, le code suivant fournit un analyseur des expressions
   arithmtiques :

   #open "genlex";;
   #open "parser";;
   let value = 
     let lexer = make_lexer ["+"; "-"; "*"; "/"; "~"; ""; "plus"; "mult"; "("; ")"; ";"]
     and syntax_of_lexme = function
         | Kwd "+" -> Infixe(1,fun x y -> x+y)
         | Kwd "-" -> Infixe(1,fun x y -> x-y)
         | Kwd "*" -> Infixe(2,fun x y -> x*y)
         | Kwd "/" -> Infixe(2,fun x y -> x/y)
         | Kwd "~" -> Prfixe(fun x -> -x)
         | Kwd "" -> Postfixe(fun x -> x*x)
         | Kwd "plus" -> Liste(fun x -> it_list (prefix +) 0 x)
         | Kwd "mult" -> Liste(fun x -> it_list (prefix * ) 1 x)
         | Int n   -> Valeur n
         | Kwd "(" -> Lpar
         | Kwd ")" -> Rpar
         | Kwd ";" -> Lsep
         | _       -> raise Parse_error
     in analyse syntax_of_lexme lexer
   ;;

   value "2 + 4*plus(1; 2; 3)/(8 + ~2)";;  (-> 26)

*)

